/*
COMS W4115
Language: YOLOP

Lisa Li
Jonathan Liu
Sasha McIntosh

*/

%{ open Ast %}

%token READ_IN WRITE_OUT SET GET DIMENSION GETC SETC
%token SEMI COMMA QUOTE LPAREN RPAREN LBRACE RBRACE LBRACK RBRACK COMMA
%token PLUS MINUS TIMES DIVIDE ASSIGN
%token EQ NEQ LT LEQ GT GEQ AND OR NOT
%token PIXEL IMAGE
%token RETURN IF ELSE FOR WHILE BREAK CONTINUE INT STRING
%token FUNCTION
%token <int> ILITERAL
%token <string> SLITERAL
%token <string> ID
%token EOF

%nonassoc NOELSE
%nonassoc ELSE

%left AND OR
%right NOT
%right ASSIGN
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS INCR DECR
%left TIMES DIVIDE

%start program
%type <Ast.program> program

%%

program:
   /* nothing */ { [], [] }
 | program vdef { ($2 :: fst $1), snd $1 }
 | program fdecl { fst $1, ($2 :: snd $1) }

fdecl:
   FUNCTION ftype_opt ID LPAREN formals_opt RPAREN LBRACE vdef_list stmt_list RBRACE
     { { fname = $3;
	 formals = $5;
	 locals = List.rev $8;
	 body = List.rev $9;
   freturn = $2} }

formals_opt:
    /* nothing */ { [] }
  | formal_list   { List.rev $1 }

ftype_opt:
/*nothing*/ {Void}
| basic_type {$1}

vdef_list:
  /* nothing */ { []}
| vdef_list vdef {$2 :: $1}

vdef:
  vdecl {VarDecl($1)}
| vinit { Varinit($1) }

vdecl:
  basic_type ID SEMI {{vname=$2;vtype=$1}}

vinit:
   basic_type ID ASSIGN expr SEMI { Vinit ({vname = $2; vtype = $1}, $4)}

basic_type:
  INT { Int }
| STRING {String}
| PIXEL {Pixel}
| IMAGE {Image}

formal_list:
   fdef	             { [$1] }
  | formal_list COMMA fdef  { $3 :: $1 }

fdef:
   basic_type ID {{ vname=$2; vtype = $1}}

stmt_list:
    /* nothing */   { [] }
  | stmt_list stmt  { $2 :: $1 }

stmt:
    expr SEMI { Expr($1) }
  | RETURN expr SEMI { Return($2) }
  | LBRACE stmt_list RBRACE { Block(List.rev $2) }
  | IF LPAREN expr RPAREN stmt %prec NOELSE { If($3, $5, Block([])) }
  | IF LPAREN expr RPAREN stmt ELSE stmt   { If($3, $5, $7) }
  | FOR LPAREN expr_opt SEMI expr_opt SEMI expr_opt RPAREN stmt  { For($3, $5, $7, $9) }
  | WHILE LPAREN expr RPAREN stmt { While($3, $5) }
  | BREAK SEMI { Break }
  | CONTINUE SEMI { Continue }


expr_opt:
    /* nothing */ { Noexpr }
  | expr          { $1 }

expr:
    SLITERAL         { StringLiteral($1)    }
  | ILITERAL         { IntLiteral($1)       }
  | ID               { Id($1)               }
  | expr PLUS   expr { Binop($1, Add,   $3) }
  | expr MINUS  expr { Binop($1, Sub,   $3) }
  | expr TIMES  expr { Binop($1, Mult,  $3) }
  | expr DIVIDE expr { Binop($1, Div,   $3) }
  | expr EQ     expr { Binop($1, Equal, $3) }
  | expr NEQ    expr { Binop($1, Neq,   $3) }
  | expr LT     expr { Binop($1, Less,  $3) }
  | expr LEQ    expr { Binop($1, Leq,   $3) }
  | expr GT     expr { Binop($1, Greater,  $3) }
  | expr GEQ    expr { Binop($1, Geq,   $3) }
  | expr AND    expr { Binop($1, And,   $3) }
  | expr OR     expr { Binop($1, Or,    $3) }
  | NOT expr         { Not($2)              }
  | LBRACK expr COMMA expr COMMA expr COMMA expr RBRACK SEMI { ImgPixel($2, $4, $6, $8) }
  | LBRACK expr COMMA expr RBRACK SEMI { ImgImg($2, $4)}
  | READ_IN LPAREN ID COMMA expr COMMA expr RPAREN SEMI { Readin($3, $5, $7)}
  | WRITE_OUT LPAREN expr COMMA expr COMMA expr RPAREN SEMI { Writeout($3, $5, $7)}
  | GET LPAREN expr COMMA expr COMMA expr RPAREN SEMI { Get($3, $5, $7)}
  | SET LPAREN expr COMMA expr COMMA expr COMMA expr RPAREN SEMI { Set($3, $5, $7, $9)}
  | GETC LPAREN expr COMMA expr RPAREN SEMI { Getc($3, $5) }
  | SETC LPAREN expr COMMA expr COMMA expr RPAREN SEMI { Setc($3, $5, $7) }
  | DIMENSION LPAREN expr COMMA SLITERAL RPAREN SEMI { Dimension($3, $5)}
  | ID ASSIGN expr  { Assign($1, $3)       }
  | ID LPAREN actuals_opt RPAREN { Call($1, $3) }
  | LPAREN expr RPAREN { $2 }

actuals_opt:
    /* nothing */ { [] }
  | actuals_list  { List.rev $1 }

actuals_list:
    expr                    { [$1] }
  | actuals_list COMMA expr { $3 :: $1 }
